#!/usr/bin/env python
# encoding: UTF-8

import base64
import binascii
import hashlib
import os
import json
import logging

from bs4 import BeautifulSoup
import requests
from Crypto.Cipher import AES

from .excs import NeteaseIOError

site_uri = 'http://music.163.com'
uri = 'http://music.163.com/api'
uri_we = 'http://music.163.com/weapi'
uri_v1 = 'http://music.163.com/weapi/v1'
uri_v3 = 'http://music.163.com/weapi/v3'
uri_e = 'https://music.163.com/eapi'

logger = logging.getLogger(__name__)


class CodeShouldBe200(NeteaseIOError):
    def __init__(self, data):
        self._code = data['code']
        self._data = data

    def __str__(self):
        return f'json code field should be 200, got {self._code}. data: {self._data}'


class API(object):
    def __init__(self):
        super().__init__()
        self.headers = {
            'Host': 'music.163.com',
            'Connection': 'keep-alive',
            'Referer': 'http://music.163.com/',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2)'
                          ' AppleWebKit/537.36 (KHTML, like Gecko)'
                          ' Chrome/33.0.1750.152 Safari/537.36'
        }
        self._cookies = dict(appver="1.2.1", os="osx")
        self._http = None

    @property
    def cookies(self):
        return self._cookies

    def load_cookies(self, cookies):
        self._cookies.update(cookies)
        # 云盘资源发布仅有似乎不支持osx平台
        self._cookies.update(dict(appver="7.2.24", os="android"))

    def set_http(self, http):
        self._http = http

    @property
    def http(self):
        return requests if self._http is None else self._http

    def request(self, method, action, query=None, timeout=2):
        # logger.info('method=%s url=%s data=%s' % (method, action, query))
        if method == "GET":
            res = self.http.get(action, headers=self.headers,
                                cookies=self._cookies, timeout=timeout)
        elif method == "POST":
            res = self.http.post(action, data=query, headers=self.headers,
                                 cookies=self._cookies, timeout=timeout)
        elif method == "POST_UPDATE":
            res = self.http.post(action, data=query, headers=self.headers,
                                 cookies=self._cookies, timeout=timeout)
            self._cookies.update(res.cookies.get_dict())
        content = res.content
        content_str = content.decode('utf-8')
        content_dict = json.loads(content_str)
        return content_dict

    def login(self, country_code, username, pw_encrypt, phone=False):
        action = 'http://music.163.com/api/login/'
        phone_action = 'http://music.163.com/api/login/cellphone/'
        data = {
            'password': pw_encrypt,
            'rememberLogin': 'true'
        }
        if username.isdigit() and (len(username) == 11 or country_code):
            phone = True
            data.update({'phone': username, 'countrycode': country_code or '86'})
        else:
            data.update({'username': username})
        if phone:
            res_data = self.request("POST_UPDATE", phone_action, data)
            return res_data
        else:
            res_data = self.request("POST_UPDATE", action, data)
            return res_data

    def check_cookies(self):
        url = uri + '/push/init'
        data = self.request("POST_UPDATE", url, {})
        if data['code'] == 200:
            return True
        return False

    def confirm_captcha(self, captcha_id, text):
        action = uri + '/image/captcha/verify/hf?id=' + str(captcha_id) +\
            '&captcha=' + str(text)
        data = self.request('GET', action)
        return data

    def get_captcha_url(self, captcha_id):
        action = 'http://music.163.com/captcha?id=' + str(captcha_id)
        return action

    def user_profile(self, user_id):
        """
        {'nickname': 'cosven',
         'avatarImg': 'xx.jpg',
         'userType': 0,
         'authStatus': 0,
         'expertTags': None,
         'backgroundUrl': 'xx.jpg',
         'playCount': 2892,
         'createdplCnt': 8,
         'starPlaylist': {...},
         'playlist': [...],
         'code': 200
        }
        """
        action = uri_we + '/share/userprofile/info'
        data = {'userId': user_id}
        payload = self.encrypt_request(data)
        res_data = self.request('POST', action, payload)
        code = res_data['code']
        if code == 200:
            return res_data
        elif code == 400:
            logger.warn(f'user:{user_id} may be invalid')
        raise CodeShouldBe200(res_data)

    # 用户歌单
    def user_playlists(self, uid, offset=0, limit=200):
        action = uri + '/user/playlist/?offset=' + str(offset) +\
            '&limit=' + str(limit) + '&uid=' + str(uid)
        data = self.request('GET', action)
        if data['code'] == 200:
            return data['playlist']
        return []

    def user_favorite_albums(self, offset=0, limit=30):
        action = uri_we + '/album/sublist'
        data = {
            'offset': offset,
            'limit': limit,
            'csrf_token': self._cookies.get('__csrf')}
        payload = self.encrypt_request(data)
        res_data = self.request('POST', action, payload)
        if res_data['code'] == 200:
            return res_data
        return None

    def user_favorite_artists(self, offset=0, limit=30):
        action = uri_we + '/artist/sublist'
        data = {
            'offset': offset,
            'limit': limit,
            'csrf_token': self._cookies.get('__csrf')}
        payload = self.encrypt_request(data)
        res_data = self.request('POST', action, payload)
        if res_data['code'] == 200:
            return res_data
        return None

    # 搜索单曲(1)，歌手(100)，专辑(10)，歌单(1000)，用户(1002) *(type)*
    def search(self, s, stype=1, offset=0, total='true', limit=30):
        """get songs list from search keywords"""
        action = uri + '/search/get'
        data = {
            's': s,
            'type': stype,
            'offset': offset,
            'total': total,
            'limit': limit
        }
        resp = self.request('POST', action, data)
        if resp['code'] == 200:
            return resp['result']
        return []

    def playlist_detail_v3(self, pid, offset=0, limit=200):
        """
        该接口返回的 ['playlist']['trackIds'] 字段会包含所有的歌曲
        """
        action = '/playlist/detail'
        url = uri_v3 + action
        data = dict(id=pid, limit=limit, offset=offset, n=limit)
        payload = self.encrypt_request(data)
        res_data = self.request('POST', url, payload)
        if res_data['code'] == 200:
            return res_data['playlist']
        raise CodeShouldBe200(res_data)

    def update_playlist_name(self, pid, name):
        url = uri + '/playlist/update/name'
        data = {
            'id': pid,
            'name': name
        }
        res_data = self.request('POST', url, data)
        return res_data

    def new_playlist(self, uid, name):
        url = uri + '/playlist/create'
        data = {
            'uid': uid,
            'name': name
        }
        res_data = self.request('POST', url, data)
        if res_data['code'] == 200:
            return res_data['playlist']
        raise CodeShouldBe200(res_data)

    def delete_playlist(self, pid):
        url = uri + '/playlist/delete'
        data = {
            'id': pid,
            'pid': pid
        }
        res_data = self.request('POST', url, data)
        if res_data['code'] == 200:
            return
        raise CodeShouldBe200(res_data)

    def artist_infos(self, artist_id):
        """
        :param artist_id: artist_id
        :return: {
            code: int,
            artist: {artist},
            more: boolean,
            hotSongs: [songs]
        }
        """
        action = uri + '/artist/' + str(artist_id)
        data = self.request('GET', action)
        return data

    def artist_songs(self, artist_id, offset=0, limit=50):
        action = uri_v1 + '/artist/songs'
        data = dict(id=artist_id,
                    limit=limit,
                    offset=offset,
                    order='hot',
                    work_type=1,
                    private_cloud='true')
        payload = self.encrypt_request(data)
        res_data = self.request('POST', action, payload)
        if res_data['code'] == 200:
            return res_data
        raise CodeShouldBe200(res_data)

    def artist_albums(self, artist_id, offset=0, limit=20):
        action = ('{uri}/artist/albums/{artist_id}?'
                  'offset={offset}&limit={limit}')
        action = action.format(uri=uri,
                               artist_id=artist_id,
                               offset=offset,
                               limit=limit)
        data = self.request('GET', action)
        return data

    # album id --> song id set
    def album_infos(self, album_id):
        """
        :param album_id:
        :return: {
            code: int,
            album: { album }
        }
        """
        action = uri + '/album/' + str(album_id)
        data = self.request('GET', action)
        if data['code'] == 200:
            return data['album']

    def album_desc(self, album_id):
        action = site_uri + '/album'
        data = {'id': album_id}
        res = self.http.get(action, data, headers=self.headers)
        if res is None:
            return None
        soup = BeautifulSoup(res.content, 'html.parser')
        albdescs = soup.select('.n-albdesc')
        if albdescs:
            return albdescs[0].prettify()
        return ''

    def artist_desc(self, artist_id):
        action = site_uri + '/artist/desc'
        data = {'id': artist_id}
        res = self.http.get(action, data, headers=self.headers)
        if res is None:
            return None
        soup = BeautifulSoup(res.content, 'html.parser')
        artdescs = soup.select('.n-artdesc')
        if artdescs:
            artdesc = artdescs[0]
            # FIXME: 艺术家描述是 html 格式的，它有一个 header 为
            # ``<h2>{artist_name}简介</h2>``, 而在 FeelUOwn 的 UI 设计中，
            # FeelUown 是把艺术家描述显示在艺术家名字下面，
            # 而艺术家名字也是用 ``<h2>{artist_name}</h2>`` 来渲染的，
            # 这样在视觉上就会出现两个非常相似的文字，非常难看，
            # 所以我们在这里把描述中的标题去掉。
            # 另外，我们还把描述中所有的 h2 header 替换成 h3 header。
            artdesc.h2.decompose()
            for h2 in artdesc.select('h2'):
                h2.name = 'h3'
            return artdesc.prettify()
        return ''

    # song id --> song url ( details )
    def song_detail(self, music_id):
        action = uri + '/song/detail/?id=' + str(music_id) + '&ids=[' +\
            str(music_id) + ']'
        data = self.request('GET', action)
        if data['code'] == 200:
            if data['songs']:
                return data['songs'][0]
            return
        raise CodeShouldBe200(data)

    def weapi_songs_url(self, music_ids, bitrate=320000):
        """
        When the expected bitrate song url does not exist, server will
        return a fallback song url. For example, we request a song
        url with bitrate=320000. If there only exists a song url with
        bitrate=128000, the server will return it.

        NOTE(cosven): After some manual testing, we found that the url is
        None in following cases:
        1. the song is for vip-user and the current user is not vip.
        2. the song is a paid song and the current user(may be a vip)
           has not bought it.
        """
        url = uri_we + '/song/enhance/player/url'
        data = {
            'ids': music_ids,
            'br': bitrate,
            'csrf_token': self._cookies.get('__csrf')
        }
        payload = self.encrypt_request(data)
        data = self.request('POST', url, payload)
        if data['code'] == 200:
            return data['data']
        return []

    def user_level(self):
        url = uri + '/user/level'
        payload = self.encrypt_request({})
        data = self.request('POST', url, payload)
        # data example
        #
        # {'userId': 12345,
        #  'info': '60G音乐网盘免费容量$黑名单上限100',
        #  'progress': 0.175,
        #  'nextPlayCount': 5000,
        #  'nextLoginCount': 150,
        #  'nowPlayCount': 875,
        #  'nowLoginCount': 150,
        #  'level': 8},
        if data['code'] == 200:
            return data['data']
        raise CodeShouldBe200(data)

    def songs_detail(self, music_ids):
        """批量获取歌曲的详细信息（老版）

        经过测试 music_ids 不能超过 200 个。
        """
        music_ids = [str(music_id) for music_id in music_ids]
        action = uri + '/song/detail?ids=[' +\
            ','.join(music_ids) + ']'
        data = self.request('GET', action)
        if data['code'] == 200:
            return data['songs']
        return []

    def songs_detail_v3(self, music_ids):
        """批量获取歌曲的详细信息

        经过测试 music_ids 不能超过 1000 个
        """
        action = '/song/detail'
        url = uri_v3 + action
        params = {
            'c': json.dumps([{'id': id_} for id_ in music_ids]),
            'ids': json.dumps(music_ids)
        }
        payload = self.encrypt_request(params)
        data = self.request('POST', url, payload)
        if data['code'] == 200:
            return data['songs']
        return []

    def op_music_to_playlist(self, mid, pid, op):
        """
        :param op: add or del
        """
        url_add = uri + '/playlist/manipulate/tracks'
        trackIds = '["' + str(mid) + '"]'
        data_add = {
            'tracks': str(mid),  # music id
            'pid': str(pid),    # playlist id
            'trackIds': trackIds,  # music id str
            'op': op   # opation
        }
        data = self.request('POST', url_add, data_add)
        code = data.get('code')

        # 从歌单中成功的移除歌曲时，code 是 200
        # 当从歌单中移除一首不存在的歌曲时，code 也是 200
        # 当向歌单添加歌曲时，如果歌曲已经在列表当中，返回 code 为 502
        # code 为 521 时，可能是因为：绑定手机号后才可操作哦
        if code == 200:
            return 1
        elif code == 502:
            return -1
        else:
            return 0

    def set_music_favorite(self, mid, flag):
        url = uri + '/song/like'
        data = {
            "trackId": mid,
            "like": str(flag).lower(),
            "time": 0
        }
        return self.request("POST", url, data)

    def get_radio_music(self):
        url = uri + '/radio/get'
        data = self.request('GET', url)
        if data['code'] == 200:
            return data['data']
        return None

    def get_mv_detail(self, mvid):
        """Get mv detail
        :param mvid: mv id
        :return:
        """
        url = uri + '/mv/detail?id=' + str(mvid)
        data = self.request('GET', url)
        if data['code'] == 200:
            return data['data']
        raise CodeShouldBe200(data)

    def get_lyric_by_songid(self, mid):
        """Get song lyric
        :param mid: music id
        :return: {
            lrc: {
                version: int,
                lyric: str
            },
            tlyric: {
                version: int,
                lyric: str
            }
            sgc: bool,
            qfy: bool,
            sfy: bool,
            transUser: {},
            code: int,
        }
        """
        # tv 表示翻译。-1：表示要翻译，1：不要
        # lv 为 1 时，对于纯音乐，API 返回的歌词为空
        # lv 为 -1 时，纯音乐返回的歌词是：纯音乐，请欣赏
        url = uri + '/song/lyric?' + 'id=' + str(mid) + '&lv=-1&kv=1&tv=-1'
        return self.request('GET', url)

    def get_similar_song(self, mid, offset=0, limit=10):
        url = (f"http://music.163.com/api/discovery/simiSong"
               f"?songid={mid}&offset={offset}&total=true&limit={limit}")
        data = self.request('GET', url)
        if data['code'] == 200:
            return data['songs']
        raise CodeShouldBe200(data)

    def get_recommend_songs(self):
        url = uri_v3 + '/discovery/recommend/songs'
        payload = self.encrypt_request({})
        res_data = self.request('POST', url, payload)
        if res_data['code'] == 200:
            return res_data['data']['dailySongs']
        raise CodeShouldBe200(res_data)

    def get_recommend_playlists(self):
        url = uri + '/discovery/recommend/resource'
        payload = self.encrypt_request({})
        data = self.request('POST', url, payload)
        if data['code'] == 200:
            return data['recommend']
        raise CodeShouldBe200(data)

    def get_comment(self, comment_id):
        data = {
            'rid': comment_id,
            'offset': '0',
            'total': 'true',
            'limit': '20',
            'csrf_token': self._cookies.get('__csrf')
        }
        url = uri_v1 + '/resource/comments/' + comment_id
        payload = self.encrypt_request(data)
        res_data = self.request('POST', url, payload)
        if res_data['code'] == 200:
            return res_data
        raise CodeShouldBe200(res_data)

    def accumulate_pl_count(self, mid):
        data = {"ids": "[%d]" % mid, "br": 128000,
                "csrf_token": self._cookies.get('__scrf')}
        url = uri_we + '/pl/count'
        payload = self.encrypt_request(data)
        return self.request('POST', url, payload)

    def cloud_songs(self, offset=0, limit=30):
        data = dict(limit=limit, offset=offset)
        url = uri_v1 + '/cloud/get'
        payload = self.encrypt_request(data)
        res_data = self.request('POST', url, payload)
        if res_data['code'] == 200:
            return res_data
        raise CodeShouldBe200(res_data)

    def cloud_songs_detail(self, music_ids):
        data = dict(songIds=music_ids.split(","))
        url = uri_v1 + '/cloud/get/byids'
        payload = self.encrypt_request(data)
        return self.request('POST', url, payload)

    def cloud_songs_delete(self, music_ids):
        data = dict(songIds=music_ids.split(","))
        url = uri_we + '/cloud/del'
        payload = self.encrypt_request(data)
        return self.request('POST', url, payload)

    def cloud_song_match(self, sid, asid):
        url = uri + f'/cloud/user/song/match?songId={sid}&adjustSongId={asid}'
        data = self.request('GET', url)
        if data['code'] == 200:
            return data['data']
        raise CodeShouldBe200(data)

    def cloud_song_upload(self, path):
        def md5sum(file):
            md5sum = hashlib.md5()
            with open(file, 'rb') as f:
                # while chunk := f.read():
                #     md5sum.update(chunk)
                md5sum.update(f.read())
            return md5sum

        from .cloud_helpers.cloud_api import Cloud_API
        cloud_api = Cloud_API(self, uri_e)

        fname = os.path.basename(path)
        fext = path.split('.')[-1]
        '''Parsing file names'''
        fsize = os.stat(path).st_size
        md5 = md5sum(path).hexdigest()
        logger.debug(f'[-] Checking file ( MD5: {md5} )')
        cresult = cloud_api.GetCheckCloudUpload(md5)
        if cresult['code'] != 200:
            return 'UPLOAD_CHECK_FAILED'

        '''网盘资源发布 4 步走：'''
        '''1.拿到上传令牌 - 需要文件名，MD5，文件大小'''
        token = cloud_api.GetNosToken(fname, md5, fsize, fext)
        if token['code'] != 200:
            return 'TOKEN_ALLOC_FAILED'
        token = token['result']

        '''2. 若文件未曾上传完毕，则完成其上传'''
        if cresult['needUpload']:
            logger.info(f'[+] {fname} needs to be uploaded ( {fsize} B )')
            try:
                upload_result = cloud_api.SetUploadObject(
                    open(path, 'rb'),
                    md5, fsize, token['objectKey'], token['token']
                )
            except Exception:
                return 'UPLOAD_FAILED'
            logger.debug(f'[-] Response:\n  {upload_result}')

        '''3. 提交资源'''
        songId = cresult['songId']
        logger.debug(f'''[!] Assuming upload has finished,preparing to submit
        ID  :   {songId}
        MD5 :   {md5}
        NAME:   {fname}''')
        metadata = cloud_api.GetMetadata(path)
        submit_result = cloud_api.SetUploadCloudInfo(
            token['resourceId'], songId, md5, fname,
            song=metadata.get('title', '.'),
            artist=metadata.get('artist', '.'),
            album=metadata.get('album', '.')
        )
        if submit_result['code'] != 200:
            return 'SUBMIT_FAILED'
        logger.debug(f'[-] Response:\n  {submit_result}')

        '''4. 发布资源'''
        publish_result = cloud_api.SetPublishCloudResource(submit_result['songId'])
        if publish_result['code'] != 200:
            return 'PUBLISH_FAILED'
        logger.debug(f'[-] Response:\n  {publish_result}')

        return 'STATUS_SUCCEEDED'

    def subscribed_djradio(self, limit=0, offset=0):
        data = dict(limit=100, time=0, needFee=False)
        url = uri_e + '/djradio/subed/v1'
        payload = self.eapi_encrypt(b'/api/djradio/subed/v1', data)
        return self.request('POST', url, {'params': payload})

    def djradio_detail(self, radio_id):
        data = dict(id=radio_id)
        url = uri_e + '/djradio/v2/get'
        payload = self.eapi_encrypt(b'/api/djradio/v2/get', data)
        return self.request('POST', url, {'params': payload})

    def djradio_song_detail(self, id_):
        data = dict(id=id_)
        url = uri_e + '/dj/program/detail'
        payload = self.eapi_encrypt(b'/api/dj/program/detail', data)
        return self.request('POST', url, {'params': payload})

    def djradio_list(self, radio_id, limit=50, offset=0, asc=False):
        data = dict(radioId=radio_id, limit=limit, offset=offset, asc=asc)
        url = uri_e + '/v1/dj/program/byradio'
        payload = self.eapi_encrypt(b'/api/v1/dj/program/byradio', data)
        return self.request('POST', url, {'params': payload})

    def list_toplist(self):
        url = uri + '/toplist'
        data = self.request('GET', url)
        if data['code'] == 200:
            return data['list']
        raise CodeShouldBe200(data)

    def _create_aes_key(self, size):
        return (''.join([hex(b)[2:] for b in os.urandom(size)]))[0:16]

    def _aes_encrypt(self, text, key):
        pad = 16 - len(text) % 16
        text = text + pad * chr(pad)
        encryptor = AES.new(bytes(key, 'utf-8'), 2, b'0102030405060708')
        enc_text = encryptor.encrypt(bytes(text, 'utf-8'))
        enc_text_encode = base64.b64encode(enc_text)
        return enc_text_encode

    def _rsa_encrypt(self, text):
        e = '010001'
        n = '00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615'\
            'bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf'\
            '695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46'\
            'bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b'\
            '8e289dc6935b3ece0462db0a22b8e7'
        reverse_text = text[::-1]
        encrypted_text = pow(int(binascii.hexlify(reverse_text), 16),
                             int(e, 16), int(n, 16))
        return format(encrypted_text, "x").zfill(256)

    def eapi_encrypt(self, path, params):
        """
        eapi接口参数加密
        :param bytes path: 请求的路径
        :param params: 请求参数
        :return str: 加密结果
        """
        params = json.dumps(params, separators=(',', ':')).encode()
        sign_src = b'nobody' + path + b'use' + params + b'md5forencrypt'
        m = hashlib.md5()
        m.update(sign_src)
        sign = m.hexdigest()
        aes_src = path + b'-36cd479b6b5-' + params + b'-36cd479b6b5-' + sign.encode()
        pad = 16 - len(aes_src) % 16
        aes_src = aes_src + bytearray([pad] * pad)
        crypt = AES.new(b'e82ckenh8dichen8', AES.MODE_ECB)
        ret = crypt.encrypt(aes_src)
        return binascii.b2a_hex(ret).upper()

    def encrypt_request(self, data):
        text = json.dumps(data)
        first_aes_key = '0CoJUm6Qyw8W8jud'
        second_aes_key = self._create_aes_key(16)
        enc_text = self._aes_encrypt(
            self._aes_encrypt(text, first_aes_key).decode('ascii'),
            second_aes_key).decode('ascii')
        enc_aes_key = self._rsa_encrypt(second_aes_key.encode('ascii'))
        payload = {
            'params': enc_text,
            'encSecKey': enc_aes_key,
        }
        return payload


api = API()


if __name__ == '__main__':
    from fuo_netease.login_controller import LoginController
    user = LoginController.load()
    cookies, _ = user.cache_get('cookies')
    api.load_cookies(cookies)
    print(api.djradio_song_detail(1883706033))



import logging

from pathlib import Path

from feeluown.gui.widgets import TextButton
from feeluown.utils import aio
from feeluown.gui.page_containers.table import Renderer

from fuo_netease import provider
from fuo_netease.qt_compat import QFileDialog, QMessageBox

logger = logging.getLogger(__name__)


async def render(req, **kwargs):
    app = req.ctx['app']
    app.ui.right_panel.set_body(app.ui.table_container)

    renderer = FavRenderer()
    await app.ui.table_container.set_renderer(renderer)


class FavRenderer(Renderer):
    def __init__(self):
        global provider

        self._user = provider._user

    async def render(self):
        self.meta_widget.show()
        self.meta_widget.title = '云盘'

        self.show_songs(await aio.run_fn(provider.current_user_cloud_songs))
        dir_upload_btn = TextButton('上传目录', self.toolbar)
        dir_upload_btn.clicked.connect(
            lambda: aio.run_afn(self._upload_cloud_songs_bydir))
        self.toolbar.add_tmp_button(dir_upload_btn)
        upload_btn = TextButton('上传音乐', self.toolbar)
        upload_btn.clicked.connect(lambda: aio.run_afn(self._upload_cloud_songs))
        self.toolbar.add_tmp_button(upload_btn)
        refresh_btn = TextButton('刷新音乐', self.toolbar)
        refresh_btn.clicked.connect(self._refresh_cloud_songs)
        self.toolbar.add_tmp_button(refresh_btn)

    async def _upload_cloud_songs_bydir(self):
        # FIXME: 目前无法根据当前页面进行自动刷新, 只能手动刷新
        # FIXME: 本地音乐还没扫描完成或歌曲播放时, 可能线程错误提示
        root = QFileDialog.getExistingDirectory(
            self.toolbar, '选择文件夹', Path.home().as_posix())
        if root == '':
            return

        import os
        paths = []
        for dir, _, files in os.walk(root):
            exts = ['mp3', 'm4a', 'wma', 'flac', 'ogg']
            paths.extend([os.path.join(dir, f)
                          for f in sorted(files)
                          if f.rsplit('.', 1)[-1] in exts])

        for idx, path in enumerate(paths):
            ok = await aio.run_fn(provider.upload_song, path)
            if ok:
                logger.warning(f'[{idx + 1}/{len(paths)}]{path} 上传成功!')
            else:
                logger.warning(f'[{idx + 1}/{len(paths)}]{path} 上传失败!')
        QMessageBox.information(self.toolbar, '上传目录', '上传完成！')

    async def _upload_cloud_songs(self):
        # FIXME: 目前无法根据当前页面进行自动刷新, 只能手动刷新
        # FIXME: 本地音乐还没扫描完成或歌曲播放时, 可能线程错误提示
        paths, _ = QFileDialog.getOpenFileNames(
            self.toolbar, '选择文件', Path.home().as_posix(),
            'Supported Files (*.mp3 *.m4a *.wma *.flac *.ogg);; All Files (*.*)')
        if not paths:
            return

        for idx, path in enumerate(paths):
            ok = await aio.run_fn(provider.upload_song, path)
            if ok:
                logger.warning(f'[{idx + 1}/{len(paths)}]{path} 上传成功!')
            else:
                logger.warning(f'[{idx + 1}/{len(paths)}]{path} 上传失败!')
        QMessageBox.information(self.toolbar, '上传音乐', '上传完成！')

    def _refresh_cloud_songs(self):
        self._app.browser.goto(page='/providers/netease/fav')



# -*- coding: utf-8 -*-

import logging

from .provider import provider

__alias__ = '网易云音乐'
__feeluown_version__ = '2.1a0'
__version__ = '0.0.2'
__desc__ = '网易云音乐'

logger = logging.getLogger(__name__)


def enable(app):
    app.library.register(provider)
    if app.mode & app.GuiMode:
        from .provider_ui import NeteaseProviderUI
        provider_ui = NeteaseProviderUI(app)
        app.pvd_ui_mgr.register(provider_ui)


def disable(app):
    app.library.deregister(provider)
    if app.mode & app.GuiMode:
        app.provider_uimgr.remove(provider.identifier)



# -*- coding: utf-8 -*-
# Miscellaneous implementations of some of  netease's security algorithms
import ctypes
import math
import json
import urllib.parse
from time import time
from . import HashDigest, HexDigest, BASE64

# region secrets
WEAPI_WATCHMAN_PAD = 5
WEAPI_WATCHMAN_SBOX = (
    "LPc4uQNptC2A6y.R90DOBIroS+qnx/eb3FM8fW1UZG7VmwvksgjhaTKzlXdiYEHJ"
)
WEAPI_WATCHMAN_KEY = [31, 125, -12, 60, 32, 48]
WEAPI_WATCHMAN_SALT_1 = "dAWsBhCqtOaNLLJ25hBzWbqWXwiK99Wd"
WEAPI_WATCHMAN_SALT_2 = "WoeTpXnDDPhiAvsJUUIY3RdAo2PKaVwi"
WEAPI_WATCHMAN_DID = "Eyj1ZVEWep5ERQEFFVMuN0xNxLvKykXT"

WEAPI_ABROAD_SBOX = [
    82,
    9,
    106,
    -43,
    48,
    54,
    -91,
    56,
    -65,
    64,
    -93,
    -98,
    -127,
    -13,
    -41,
    -5,
    124,
    -29,
    57,
    -126,
    -101,
    47,
    -1,
    -121,
    52,
    -114,
    67,
    68,
    -60,
    -34,
    -23,
    -53,
    84,
    123,
    -108,
    50,
    -90,
    -62,
    35,
    61,
    -18,
    76,
    -107,
    11,
    66,
    -6,
    -61,
    78,
    8,
    46,
    -95,
    102,
    40,
    -39,
    36,
    -78,
    118,
    91,
    -94,
    73,
    109,
    -117,
    -47,
    37,
    114,
    -8,
    -10,
    100,
    -122,
    104,
    -104,
    22,
    -44,
    -92,
    92,
    -52,
    93,
    101,
    -74,
    -110,
    108,
    112,
    72,
    80,
    -3,
    -19,
    -71,
    -38,
    94,
    21,
    70,
    87,
    -89,
    -115,
    -99,
    -124,
    -112,
    -40,
    -85,
    0,
    -116,
    -68,
    -45,
    10,
    -9,
    -28,
    88,
    5,
    -72,
    -77,
    69,
    6,
    -48,
    44,
    30,
    -113,
    -54,
    63,
    15,
    2,
    -63,
    -81,
    -67,
    3,
    1,
    19,
    -118,
    107,
    58,
    -111,
    17,
    65,
    79,
    103,
    -36,
    -22,
    -105,
    -14,
    -49,
    -50,
    -16,
    -76,
    -26,
    115,
    -106,
    -84,
    116,
    34,
    -25,
    -83,
    53,
    -123,
    -30,
    -7,
    55,
    -24,
    28,
    117,
    -33,
    110,
    71,
    -15,
    26,
    113,
    29,
    41,
    -59,
    -119,
    111,
    -73,
    98,
    14,
    -86,
    24,
    -66,
    27,
    -4,
    86,
    62,
    75,
    -58,
    -46,
    121,
    32,
    -102,
    -37,
    -64,
    -2,
    120,
    -51,
    90,
    -12,
    31,
    -35,
    -88,
    51,
    -120,
    7,
    -57,
    49,
    -79,
    18,
    16,
    89,
    39,
    -128,
    -20,
    95,
    96,
    81,
    127,
    -87,
    25,
    -75,
    74,
    13,
    45,
    -27,
    122,
    -97,
    -109,
    -55,
    -100,
    -17,
    -96,
    -32,
    59,
    77,
    -82,
    42,
    -11,
    -80,
    -56,
    -21,
    -69,
    60,
    -125,
    83,
    -103,
    97,
    23,
    43,
    4,
    126,
    -70,
    119,
    -42,
    38,
    -31,
    105,
    20,
    99,
    85,
    33,
    12,
    125,
]
WEAPI_ABROAD_KEY = "fuck~#$%^&*(458"  # ookay...
WEAPI_ABROAD_IV = ([ord(s) for s in WEAPI_ABROAD_KEY] * 5)[:64]
# `abroad` initalization vector, also acts as its key
# endregion


# region functions to mimic the defficiencies of JS bit operations
def jint(v):
    return ctypes.c_int(v).value


# Bit shifting only takes care of LS 5 bits
def jls(v, bs):
    return jint(v << (bs & 0x1F))


def jrs(v, bs):
    return jint(v >> (bs & 0x1F))


# some bitwise operations are only done on the least significant 32 bits
def jmask(v):
    return v & 0xFFFFFFFF


def jxor(v1, v2):
    return jint(jmask(v1) ^ jmask(v2))


# endregion

# region common practices
def string_to_charcodes(e):
    # String to charcode array
    return [ord(char) for char in wm_encode(e)] if isinstance(e, str) else e


def char_to_hex(e):
    # Char to hexdecimal [00,FF] string
    c = "0123456789abcdef"
    return c[e >> 4 & 15] + c[e & 15]


def to_hex_string(e):
    # int array to hex string
    return "".join([char_to_hex(a) for a in e])


def cast_to_signed(e):
    # casts any integer value to signed char
    if e < -128:
        return cast_to_signed(e + 256)
    if e >= -128 and e <= 127:
        return e
    if e > 127:
        return cast_to_signed(e - 256)


def cast_to_multi_signed(b):
    # casts int32 to multiple signed chars
    return [
        cast_to_signed(b >> 24 & 255),
        cast_to_signed(b >> 16 & 255),
        cast_to_signed(b >> 8 & 255),
        cast_to_signed(b & 255),
    ]


# endregion

# region Watchman.js security
def wm_Hash(e):
    # MD5 hasher
    return HashDigest(e)


def wm_HexHash(e):
    # digests md5 hash to hex string
    return HexDigest(e)


def wm_encode(e):
    # encodeURIComponent() in watchman.js. not necessary though.
    return e


def wm_unescape(e):
    # since there's nothing escaped anyway...
    return e


def wm_xor_cipher_La(e):
    # Simple XOR cipherer
    d = WEAPI_WATCHMAN_KEY
    e = string_to_charcodes(e)
    g = []
    for q in range(0, len(e)):
        g.append(cast_to_signed(jxor(e[q], d[q % len(d)])))
        g[q] = cast_to_signed(-g[q])
    return to_hex_string(g)


def wm_pbox_ga(e, c, d, r, g):
    # p-box
    for h in range(0, g):
        d[r + h] = e[c + h]


def wm_hex_to_ints_hb(e):
    # Hexstring to integer array
    c = []
    for r in range(0, len(e), 2):
        h = jls(int(e[r], 16), 4)
        m = int(e[r + 1], 16)
        c.append(cast_to_signed(h + m))
    return c


def wm_sbox_Qb(e, c, d=0, r=WEAPI_WATCHMAN_SBOX, g=WEAPI_WATCHMAN_PAD):
    # s-box
    h, m, g = 0, [], str(g)

    def push(*a):
        m.extend(a)

    if d == 1:
        d = e[c]
        h = 0
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], g, g)
    elif d == 2:
        d = e[c]
        h = e[c + 1]
        e = 0
        push(
            r[d >> 2 & 63],
            r[(d << 4 & 48) + (h >> 4 & 15)],
            r[(h << 2 & 60) + (e >> 6 & 3)],
            g,
        )
    elif d == 3:
        d = e[c]
        h = e[c + 1]
        e = e[c + 2]
        push(
            r[d >> 2 & 63],
            r[(d << 4 & 48) + (h >> 4 & 15)],
            r[(h << 2 & 60) + (e >> 6 & 3)],
            r[e & 63],
        )
    else:
        return None
    return "".join(m)


def wm_apply_sbox_Sb(e, c=[], d=WEAPI_WATCHMAN_PAD):
    # applies sbox to input e
    r = 3
    h = 0
    g = []
    while h < len(e):
        if h + r <= len(e):
            g.append(wm_sbox_Qb(e, h, r, c, d))
            h += r
        else:
            g.append(wm_sbox_Qb(e, h, len(e) - h, c, d))
            break
    return "".join(g)


def wm_map_string_Tb(a):
    return wm_apply_sbox_Sb(a, BASE64, "=")


def wm_generate_OTP_b():
    # Tokenzies current time
    e = int(time() * 1000)
    c = math.floor(e / 4294967296)
    d = e % 4294967296
    e = cast_to_multi_signed(c)
    d = cast_to_multi_signed(d)
    c = [0] * 8
    wm_pbox_ga(e, 0, c, 0, 4)
    wm_pbox_ga(d, 0, c, 4, 4)
    d = [0] * 8  # should be random nums of [0-255] * 8
    e = [0] * 16
    for g in range(0, len(c) * 2):
        if g % 2 == 0:
            f = int(g / 2)
            e[g] = (
                e[g]
                | (d[f] & 16) >> 4
                | (d[f] & 32) >> 3
                | (d[f] & 64) >> 2
                | (d[f] & 128) >> 1
                | (c[f] & 16) >> 3
                | (c[f] & 32) >> 2
                | (c[f] & 64) >> 1
                | (c[f] & 128) >> 0
            )
        else:
            f = math.floor(g / 2)
            e[g] = (
                e[g]
                | (d[f] & 1) << 0
                | (d[f] & 2) << 1
                | (d[f] & 4) << 2
                | (d[f] & 8) << 3
                | (c[f] & 1) << 1
                | (c[f] & 2) << 2
                | (c[f] & 4) << 3
                | (c[f] & 8) << 4
            )
        e[g] = cast_to_signed(e[g])
    c = to_hex_string(e)
    c = wm_HexHash(wm_Hash(wm_unescape(wm_encode(c + WEAPI_WATCHMAN_SALT_1))))
    c = wm_hex_to_ints_hb(c[:16])
    c = wm_map_string_Tb(c + e)
    return c


def wm_generate_config_chiper_bc(c, f=1, g=WEAPI_WATCHMAN_DID):
    # Generates checkToken pararmeter with current time-of-day as argument
    t = wm_unescape(wm_encode(str(f) + str(g) + str(c) + WEAPI_WATCHMAN_SALT_2))
    t = wm_Hash(t)
    return wm_xor_cipher_La(
        json.dumps(
            {
                "r": f,  # config.va - should be 1 all the time
                "d": g,  # g || "",device ID
                "b": c,
                "t": wm_map_string_Tb(t),
            }
        )
    )


# endregion

# region core.js security
def c_quote_int_as_hex(a):
    return "%" + "%".join([char_to_hex(i) for i in a])


def c_signed_xor(v1, v2):
    return jxor(cast_to_signed(v1), cast_to_signed(v2))


def c_apply_sbox(src, map):
    return [map[(i >> 4 & 15) * 16 + (i & 15)] for i in src]


def c_decrypt_abroad_message(hexstring):
    # Decrypts 'result' parameter when 'abroad' flag is true in responses
    source = wm_hex_to_ints_hb(hexstring)
    result = []
    boxA = WEAPI_ABROAD_IV
    for i in range(0, len(source), 64):
        box = source[i : i + 64]
        boxB = c_apply_sbox(
            c_apply_sbox(box, WEAPI_ABROAD_SBOX), WEAPI_ABROAD_SBOX
        )
        boxC = [c_signed_xor(boxB[i], boxA[i]) for i in range(0, 64)]
        boxD = [
            cast_to_signed(boxC[i] + cast_to_signed(-boxA[i]))
            for i in range(0, 64)
        ]
        boxE = [c_signed_xor(boxD[i], WEAPI_ABROAD_IV[i]) for i in range(0, 64)]
        boxA = source[i : i + 64]  # S-box operation, getting ready for next one
        result += boxE
    # last 4 bytes contains length
    # don't need them for now as we can just strip them away
    result = c_quote_int_as_hex(result[:-4])
    result = urllib.parse.unquote(result).strip("\x00")
    return result


# endregion



"""

关于命名
~~~~~~~~~~~~~~~
V2{X}Schema 返回的 model 都是 v2 版本的 Model 实例，也就是从 feeluown.library
模块导入进来的 Model 类。

{X}SchemaForV3 的数据来源都是 api uri_v3 接口获取的数据。
"""

import logging
from datetime import datetime

from marshmallow import Schema, post_load, fields, EXCLUDE

from feeluown.library import (
    SongModel, BriefAlbumModel, BriefArtistModel, ModelState, BriefSongModel,
    VideoModel, AlbumModel, ArtistModel, PlaylistModel, BriefUserModel,
    SimpleSearchResult, MediaFlags, AlbumType
)
from feeluown.media import Quality, MediaType, Media

logger = logging.getLogger(__name__)


class BaseSchema(Schema):
    source = fields.Str(load_default='netease')

    class Meta:
        unknown = EXCLUDE


Schema = BaseSchema
Unknown = 'Unknown'
DjradioPrefix = 'djradio_'


def create_model(model_cls, data, fields_to_cache=None):
    """
    maybe this function should be provided by feeluown

    :param fields_to_cache: list of fields name to be cached
    """
    if fields_to_cache is not None:
        cache_data = {}
        for field in fields_to_cache:
            value = data.pop(field)
            if value is not None:
                cache_data[field] = value
        model = model_cls(**data)
        for field, value in cache_data.items():
            model.cache_set(field, value)
    else:
        model = model_cls(**data)
    return model


def get_media_flags(fee: int):
    # 0: 免费或无版权
    # 1: VIP 歌曲
    # 4: 购买专辑
    # 8: 非会员可免费播放低音质，会员可播放高音质及下载
    if fee == 1:
        return MediaFlags.vip
    elif fee == 4:
        return MediaFlags.pay
    return None


class V2MvSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    title = fields.Str(required=True, data_key='name')
    cover = fields.Str(required=True)
    brs = fields.Dict(required=True)
    artists = fields.List(fields.Nested('V2BriefArtistSchema'))
    duration = fields.Int(required=True)

    @post_load
    def create_model(self, data, **kwargs):
        brs = data.pop('brs')
        q_media_mapping = {}
        for q, url in brs.items():
            media = Media(url, type_=MediaType.video)
            if q == '1080':
                quality = Quality.Video.fhd
            elif q == '720':
                quality = Quality.Video.hd
            elif q == '480':
                quality = Quality.Video.sd
            elif q == '240':
                quality = Quality.Video.ld
            else:
                logger.warning('There exists another quality:%s mv.', q)
                quality = Quality.Video.sd
            q_media_mapping[quality] = media
        data['q_media_mapping'] = q_media_mapping
        data['identifier'] = 'mv_' + str(data['identifier'])
        return create_model(VideoModel, data, ['q_media_mapping'])


class V2BriefAlbumSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str(required=True, allow_none=True)
    # cover = fields.Str(data_key='picUrl', allow_none=True)
    artist = fields.Dict()

    @post_load
    def create_v2_model(self, data, **kwargs):
        if data['name'] is None:
            data['name'] = Unknown
        if 'artist' in data:
            artist = data.pop('artist')
            data['artists_name'] = artist['name']
        return BriefAlbumModel(**data)


class V2BriefArtistSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str(required=True, allow_none=True)

    # cover = fields.Str(data_key='picUrl', allow_none=True)
    # songs = fields.List(fields.Nested('V2SongSchema'))

    @post_load
    def create_v2_model(self, data, **kwargs):
        if data['name'] is None:
            data['name'] = Unknown
        return BriefArtistModel(**data)


class V2SongSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    title = fields.Str(required=True, data_key='name', allow_none=True)
    duration = fields.Float(required=True)
    album = fields.Nested('V2BriefAlbumSchema')
    artists = fields.List(fields.Nested('V2BriefArtistSchema'))

    mv_id = fields.Int(required=True, data_key='mvid')
    comment_thread_id = fields.Str(data_key='commentThreadId',
                                   allow_none=True, load_default=None)
    fee = fields.Int(required=True)

    @post_load
    def create_v2_model(self, data, **kwargs):
        # https://github.com/feeluown/FeelUOwn/issues/499
        if data['title'] is None:
            data['title'] = Unknown

        fee = data.pop('fee')
        flags = get_media_flags(fee)
        if flags is not None:
            data['media_flags'] = flags
        return create_model(SongModel, data, ['mv_id', 'comment_thread_id'])


class V2SongSchemaForV3(Schema):
    identifier = fields.Int(required=True, data_key='id')
    title = fields.Str(required=True, data_key='name')
    duration = fields.Float(required=True, data_key='dt')
    album = fields.Nested('V2BriefAlbumSchema', data_key='al', allow_none=True)
    artists = fields.List(fields.Nested('V2BriefArtistSchema'),
                          data_key='ar',
                          allow_none=True)

    mv_id = fields.Int(required=True, data_key='mv')
    fee = fields.Int(required=True)

    @post_load
    def create_v2_model(self, data, **kwargs):
        data['artists'] = data['artists'] or []

        fee = data.pop('fee')
        flags = get_media_flags(fee)
        if flags is not None:
            data['media_flags'] = flags
        return create_model(SongModel, data, ['mv_id'])


class V2AlbumSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str(required=True)
    cover = fields.Str(data_key='picUrl', allow_none=True)
    artists = fields.List(fields.Nested('V2BriefArtistSchema'))
    # 收藏和搜索接口返回的 album 数据中的 songs 为 None
    songs = fields.List(fields.Nested('V2SongSchema'), allow_none=True)
    song_count = fields.Int(data_key='size')

    # Description is fetched seperatelly by `album_desc` API.
    description = fields.Str(load_default='')
    released = fields.Int(data_key='publishTime', load_default=0)

    # Single/专辑/"EP/Single"/合集/专辑/精选集
    type = fields.Str(required=True, data_key='type')

    @post_load
    def create_v2_model(self, data, **kwargs):
        released = data['released']
        if released:
            released_date = datetime.fromtimestamp(released / 1000)
            released_str = released_date.strftime('%Y-%m-%d')
            data['released'] = released_str
        data['songs'] = data['songs'] or []
        type_str = data.pop('type')
        if type_str == 'Single':
            type_ = AlbumType.single
        elif type_str == 'EP/Single':
            type_ = AlbumType.ep
        elif type_str == '合集':
            type_ = AlbumType.compilation
        elif type_str == '精选集':
            type_ = AlbumType.retrospective
        else:
            type_ = AlbumType.standard
        data['type_'] = type_
        return AlbumModel(**data)


class V2ArtistSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str()
    pic_url = fields.Str(data_key='picUrl', allow_none=True)
    hot_songs = fields.List(fields.Nested('V2SongSchema'), data_key='songs')

    # TODO:
    aliases = fields.List(fields.Str(), load_default=[])

    # Description is fetched seperatelly by `artist_desc` API.
    description = fields.Str(load_default='')

    @post_load
    def create_v2_model(self, data, **kwargs):
        return ArtistModel(**data)


class NAlbumSchemaV3(Schema):
    # 如果 album 无效，id 则为 0
    # 只有当 album 无效时，name 才可能为 None
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str(required=True, allow_none=True)

    @post_load
    def create_model(self, data, **kwargs):
        album = BriefAlbumModel(**data)
        if album.identifier == 0:
            album.state = ModelState.not_exists
            album.name = ''
        return album


class NeteaseDjradioSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    name = fields.Str(required=True)
    description = fields.Str(data_key='desc', required=False)
    cover = fields.Str(required=False, data_key='picUrl')

    @post_load
    def create_model(self, data, **kwargs):
        identifier = data['identifier']
        data['identifier'] = f'{DjradioPrefix}{identifier}'
        # TODO: set creator properly.
        return PlaylistModel(**data, creator=None)


class NDjradioSchema(Schema):
    # identifier = fields.Int(required=True, data_key='id')
    # title = fields.Str(required=True, data_key='name')
    main_song = fields.Dict(required=True, data_key='mainSong')
    # cover = fields.Str(required=False, data_key='coverUrl')
    radio = fields.Dict(required=True, data_key='radio')

    @post_load
    def create_model(self, data, **kwargs):

        def to_duration_ms(duration):
            seconds = duration / 1000
            m, s = seconds / 60, seconds % 60
            return '{:02}:{:02}'.format(int(m), int(s))

        song = data.pop('main_song')
        album_name = data.pop('radio')['name']
        artists_name = ','.join([artist['name'] for artist in song['artists']])

        return BriefSongModel(identifier=song['id'],
                              title=song['name'],
                              duration_ms=to_duration_ms(song['duration']),
                              artists_name=artists_name,
                              album_name=album_name,
                              state=ModelState.cant_upgrade,
                              **data)


class NCloudSchema(Schema):
    main_song = fields.Dict(required=True, data_key='simpleSong')
    album_name = fields.Str(required=True, data_key='album')
    artists_name = fields.Str(required=True, data_key='artist')

    @post_load
    def create_model(self, data, **kwargs):

        def to_duration_ms(duration):
            seconds = duration / 1000
            m, s = seconds / 60, seconds % 60
            return '{:02}:{:02}'.format(int(m), int(s))

        song = data.pop('main_song')

        return BriefSongModel(identifier=song['id'],
                              title=song['name'],
                              duration_ms=to_duration_ms(song['dt']),
                              state=ModelState.cant_upgrade,
                              **data)


class V2PlaylistCreatorScehma(Schema):
    identifier = fields.Int(required=True, data_key='userId')
    name = fields.Str(required=True, data_key='nickname')

    @post_load
    def create_v2_model(self, data, **kwargs):
        data['identifier'] = str(data['identifier'])
        return BriefUserModel(**data)


class V2PlaylistSchema(Schema):
    identifier = fields.Int(required=True, data_key='id')
    creator = fields.Nested(V2PlaylistCreatorScehma, load_default=None)
    name = fields.Str(required=True)
    description = fields.Str(required=True, allow_none=True, data_key='description')
    cover = fields.Url(required=True, data_key='coverImgUrl')

    @post_load
    def create_v2_model(self, data, **kwargs):
        if data.get('description') is None:
            data['description'] = ''
        return PlaylistModel(**data)


class NeteaseSearchSchema(Schema):
    """搜索结果 Schema"""
    q = fields.Str()
    songs = fields.List(fields.Nested(V2SongSchema))
    albums = fields.List(fields.Nested(V2AlbumSchema))
    artists = fields.List(fields.Nested(V2BriefArtistSchema))
    playlists = fields.List(fields.Nested(V2PlaylistSchema))

    @post_load
    def create_model(self, data, **kwargs):
        data.pop('source')
        return SimpleSearchResult(**data)



import os
import requests
import json
import logging

from . import security

BUCKET = "jd-musicrep-privatecloud-audio-public"

logger = logging.getLogger(__name__)


def GenerateCheckToken():
    """Generates `checkToken` parameter ,which is needed by a handful of Weapis"""
    return security.wm_generate_config_chiper_bc(security.wm_generate_OTP_b())


class Cloud_API:
    def __init__(self, api, uri_e):
        self.api = api
        self.uri_e = uri_e

    def GetMetadata(self, fpath):
        from mutagen import MutagenError
        from mutagen.mp3 import EasyMP3
        from mutagen.easymp4 import EasyMP4
        from mutagen.flac import FLAC
        from mutagen.apev2 import APEv2

        try:
            if (
                fpath.endswith("mp3")
                or fpath.endswith("ogg")
                or fpath.endswith("wma")
            ):
                audio = EasyMP3(fpath)
            elif (
                fpath.endswith("m4a")
                or fpath.endswith("m4v")
                or fpath.endswith("mp4")
            ):
                audio = EasyMP4(fpath)
            elif fpath.endswith("flac"):
                audio = FLAC(fpath)
            elif fpath.endswith("ape"):
                audio = APEv2(fpath)
            elif fpath.endswith("wav"):
                audio = dict()
        except MutagenError as e:
            logger.warning(
                "Mutagen parse metadata failed, ignore.\n"
                "file: {}, exception: {}".format(fpath, str(e))
            )
            return None

        metadata_dict = dict(audio)
        for key in metadata_dict.keys():
            metadata_dict[key] = metadata_dict[key][0]
        if "title" not in metadata_dict:
            title = os.path.split(fpath)[-1].split(".")[0]
            metadata_dict["title"] = title
        return metadata_dict

    def GetCheckCloudUpload(
        self, md5, ext="", length=0, bitrate=0, songId=0, version=1
    ):
        """移动端 - 检查云盘资源

        Args:
            md5 (str): 资源MD5哈希
            ext (str, optional): 文件拓展名. Defaults to ''.
            length (int, optional): 文件大小. Defaults to 0.
            bitrate (int, optional): 音频 - 比特率. Defaults to 0.
            songId (int, optional): 云盘资源ID. Defaults to 0 表示新资源.
            version (int, optional): 上传版本. Defaults to 1.

        Returns:
            dict
        """
        data = {
            "songId": str(songId),
            "version": str(version),
            "md5": str(md5),
            "length": str(length),
            "ext": str(ext),
            "bitrate": str(bitrate),
            "checkToken": GenerateCheckToken(),
        }
        url = self.uri_e + "/cloud/upload/check"
        payload = self.api.eapi_encrypt(b"/api/cloud/upload/check", data)
        return self.api.request("POST", url, {"params": payload})

    def GetNosToken(
        self,
        filename,
        md5,
        fileSize,
        ext,
        type="audio",
        nos_product=3,
        bucket=BUCKET,
        local=False,
    ):
        """移动端 - 云盘占位

        Args:
            filename (str): 文件名
            md5 (str): 文件 MD5
            fileSize (str): 文件大小
            ext (str): 文件拓展名
            type (str, optional): 上传类型. Defaults to 'audio'.
            nos_product (int, optional): APP类型. Defaults to 3.
            bucket (str, optional): 转存bucket.
                Defaults to 'jd-musicrep-privatecloud-audio-public'.
            local (bool, optional): 未知. Defaults to False.

        Returns:
            dict
        """
        data = {
            "type": str(type),
            "nos_product": str(nos_product),
            "md5": str(md5),
            "local": str(local).lower(),
            "filename": str(filename),
            "fileSize": str(fileSize),
            "ext": str(ext),
            "bucket": str(bucket),
            "checkToken": GenerateCheckToken(),
        }
        url = self.uri_e + "/nos/token/alloc"
        payload = self.api.eapi_encrypt(b"/api/nos/token/alloc", data)
        return self.api.request("POST", url, {"params": payload})

    def SetUploadObject(
        self,
        stream,
        md5,
        fileSize,
        objectKey,
        token,
        offset=0,
        compete=True,
        bucket=BUCKET,
    ):
        """移动端 - 上传内容

        Args:
            stream : bytes / File 等数据体 .e.g open('file.mp3')
            md5 : 数据体哈希
            objectKey : GetNosToken 获得
            token : GetNosToken 获得
            offset (int, optional): 续传起点. Defaults to 0.
            compete (bool, optional): 文件是否被全部上传. Defaults to True.

        Returns:
            dict
        """
        r = requests.post(
            "http://45.127.129.8/%s/" % bucket + objectKey.replace("/", "%2F"),
            data=stream,
            params={
                "version": "1.0",
                "offset": offset,
                "complete": str(compete).lower(),
            },
            headers={
                "x-nos-token": token,
                "Content-MD5": md5,
                "Content-Type": "cloudmusic",
                "Content-Length": str(fileSize),
            },
        )
        return json.loads(r.text)

    def SetUploadCloudInfo(
        self, resourceId, songid, md5, filename, song=".", artist=".", album="."
    ):
        """移动端 - 云盘资源提交

        注：
            - MD5 对应文件需已被 SetUploadObject 上传
            - song 项不得包含字符 .和/

        Args:
            resourceId (str): GetNosToken 获得
            songid (str): GetCheckCloudUpload 获得
            md5 (str): 文件MD5哈希
            filename (str): 文件名
            song (str, optional): 歌名 / 标题. Defaults to ''.
            artist (str, optional): 艺术家名. Defaults to ''.
            album (str, optional): 专辑名. Defaults to ''.

        WIP - 封面ID,歌词ID 等

        Returns:
            dict
        """
        data = {
            "resourceId": str(resourceId),
            "songid": str(songid),
            "md5": str(md5),
            "filename": str(filename),
            "song": str(song),
            "artist": str(artist),
            "album": str(album),
        }
        url = self.uri_e + "/upload/cloud/info/v2"
        payload = self.api.eapi_encrypt(b"/api/upload/cloud/info/v2", data)
        return self.api.request("POST", url, {"params": payload})

    def SetPublishCloudResource(self, songid):
        """移动端 - 云盘资源发布

        Args:
            songid (str): 来自 SetUploadCloudInfo

        Returns:
            SetUploadCloudInfo
        """
        data = {"songid": str(songid), "checkToken": GenerateCheckToken()}
        url = self.uri_e + "/cloud/pub/v2"
        payload = self.api.eapi_encrypt(b"/api/cloud/pub/v2", data)
        return self.api.request("POST", url, {"params": payload})



import os
import importlib

# 新版 FeelUOwn 会在启动的时候，设置这个环境变量
qt_moduel_name = os.getenv('FEELUOWN_QT_API', 'pyqt5')

if qt_moduel_name == 'pyqt6':
    qt_moduel_name = 'PyQt6'
else:
    qt_moduel_name = 'PyQt5'

QtCore = importlib.import_module(qt_moduel_name + ".QtCore", package=qt_moduel_name)
QtGui = importlib.import_module(qt_moduel_name + ".QtGui", package=qt_moduel_name)
QtWidgets = importlib.import_module(qt_moduel_name + ".QtWidgets", package=qt_moduel_name)

QLineEdit = QtWidgets.QLineEdit
QLabel = QtWidgets.QLabel
QPushButton = QtWidgets.QPushButton
QDialog = QtWidgets.QDialog
QVBoxLayout = QtWidgets.QVBoxLayout
QFileDialog = QtWidgets.QFileDialog
QMessageBox = QtWidgets.QMessageBox

pyqtSignal = QtCore.pyqtSignal

if qt_moduel_name == 'pyqt5':
    RichText = QtCore.Qt.RichText
    Password = QLineEdit.Password
else:
    RichText = QtCore.Qt.TextFormat.RichText
    Password = QLineEdit.EchoMode.Password




import json
import logging
import os
from typing import Optional

from feeluown.library import BriefUserModel
from .api import api
from .consts import USERS_INFO_FILE

logger = logging.getLogger(__name__)


def create_user(identifier, name, cookies):
    user = BriefUserModel(
        source='netease',
        identifier=identifier,
        name=name,
    )
    user.cache_set('cookies', cookies)
    return user


class LoginController(object):
    _api = api

    def __init__(self, username, uid, name, img):
        super().__init__()
        self.username = username
        self.uid = uid
        self.name = name

    @classmethod
    def create(cls, data):
        user_data = data['data']
        # img = user_data['profile']['avatarUrl']
        uid = user_data['profile']['userId']
        name = user_data['profile']['nickname']
        user = create_user(uid, name, cls._api.cookies)
        return user

    @classmethod
    def check(cls, country_code, username, pw):
        data = cls._api.login(country_code, username, pw)
        if data is None:
            return {'code': 408, 'message': '网络状况不好'}
        elif data['code'] == 200:
            if data.get("profile") is None:
                return {'code': 302, 'message': '请先在网页版绑定手机'}
            return {'code': 200, 'message': '登陆成功',
                    'data': data, 'username': username}
        elif data['code'] == 415:
            captcha_id = data['captchaId']
            url = cls._api.get_captcha_url(captcha_id)
            return {'code': 415, 'message': '本次登陆需要验证码',
                    'captcha_url': url, 'captchar_id': captcha_id}
        elif data['code'] == 501:
            return {'code': 501, 'message': '用户名不存在'}
        elif data['code'] == 502:
            return {'code': 502, 'message': '密码错误'}
        elif data['code'] == 509:
            return {'code': 509, 'message': '请休息几分钟再尝试'}
        else:
            return {'code': data['code'], 'message': data['message']}

    @classmethod
    def check_captcha(cls, captcha_id, text):
        flag, cid = cls._api.confirm_captcha(captcha_id, text)
        if flag is not True:
            url = cls._api.get_captcha_url(cid)
            return {'code': 415, 'message': '验证码错误',
                    'captcha_url': url, 'captcha_id': cid}
        return {'code': 200, 'message': '验证码正确'}

    @classmethod
    def save(cls, user):
        with open(USERS_INFO_FILE, 'w+') as f:
            data = {
                user.name: {
                    'uid': user.identifier,
                    'name': user.name,
                    'cookies': user.cache_get('cookies')[0]
                }
            }
            if f.read() != '':
                data.update(json.load(f))
            json.dump(data, f, indent=4)

    @classmethod
    def load(cls) -> Optional[BriefUserModel]:
        if not os.path.exists(USERS_INFO_FILE):
            return None
        with open(USERS_INFO_FILE, 'r') as f:
            text = f.read()
            if text == '':
                return None
            data = json.loads(text)
            username = next(iter(data.keys()))
            user_data = data[username]
            uid = user_data['uid']
            name = user_data['name']
            cookies = user_data.get('cookies', cls._api.cookies)
            user = create_user(uid, name, cookies)
        return user



import logging
from typing import List

from feeluown.library import AbstractProvider, ProviderV2, ProviderFlags as PF, \
    CommentModel, BriefCommentModel, BriefUserModel, UserModel, SongModel, \
    NoUserLoggedIn, LyricModel, ModelNotFound
from feeluown.media import Quality, Media
from feeluown.library import ModelType, SearchType
from feeluown.utils.dispatch import Signal
from feeluown.utils.reader import create_reader, SequentialReader
from .api import API, CodeShouldBe200
from .login_controller import LoginController
from .excs import NeteaseIOError


logger = logging.getLogger(__name__)
SOURCE = 'netease'


class NeteaseProvider(AbstractProvider, ProviderV2):
    class meta:
        identifier = SOURCE
        name = '网易云音乐'
        # TODO: remove
        flags = {
            ModelType.song: (PF.model_v2 | PF.similar | PF.multi_quality |
                             PF.get | PF.hot_comments | PF.web_url |
                             PF.lyric | PF.mv),
            ModelType.album: (PF.model_v2 | PF.get),
            ModelType.artist: (PF.model_v2 | PF.get | PF.songs_rd | PF.albums_rd),
            ModelType.video: (PF.model_v2 | PF.get | PF.multi_quality),
            ModelType.playlist: (PF.model_v2 | PF.get |
                                 PF.songs_rd | PF.add_song | PF.remove_song),
            ModelType.none: PF.current_user,
        }

    def __init__(self):
        super().__init__()
        self.api = API()
        self.current_user_changed = Signal()

    @property
    def identifier(self):
        return SOURCE

    @property
    def name(self):
        return '网易云音乐'

    def auto_login(self):
        logger.info("Try auto logging...")
        user = LoginController.load()
        if user is not None:
            self.auth(user)
            logger.info(f"Auto logging ok, user: {user.name}")
        else:
            logger.info("Auto logging failed, no user found.")

    def auth(self, user):
        cookies, exists = user.cache_get('cookies')
        assert exists and cookies is not None
        self._user = user
        self.api.load_cookies(cookies)
        self.current_user_changed.emit(user)

    def has_current_user(self):
        return self._user is not None

    def get_current_user(self):
        if self._user is None:
            raise NoUserLoggedIn
        # 理论上，self._user 的 cookies 缓存总是存在的。
        # 因为在创建 user 的各个逻辑里面，cookies 字段是带上的。
        if isinstance(self._user, UserModel):
            return self._user
        user = self.user_get(self._user.identifier)
        cookies, exists = self._user.cache_get('cookies')
        if exists:
            user.cache_set('cookies', cookies)
        return user

    def user_get(self, identifier):
        data = self.api.user_profile(identifier)
        user = UserModel(identifier=str(identifier),
                         source=SOURCE,
                         name=data['nickname'],
                         avatar_url=data['avatarImg'])
        return user

    def current_user_fav_create_playlists_rd(self):
        _, fav_playlists = self.current_user_playlists()
        return (
            fav_playlists +
            list(create_g(self.api.subscribed_djradio, NeteaseDjradioSchema, 'djRadios'))
        )

    def current_user_fav_create_artists_rd(self):
        return create_g(self.api.user_favorite_artists, V2BriefArtistSchema)

    def current_user_fav_create_albums_rd(self):
        return create_g(self.api.user_favorite_albums, V2BriefAlbumSchema)

    def current_user_fav_create_songs_rd(self):
        if not self.has_current_user():
            raise NoUserLoggedIn

        playlists = self.current_user_list_playlists()
        if not playlists:
            return create_reader(())

        fav_playlist = playlists[0]
        return self.playlist_create_songs_rd(fav_playlist)

    def current_user_cloud_songs(self):
        return create_cloud_songs_g(
            self.api.cloud_songs,
            self.api.cloud_songs_detail,
            V2SongSchemaForV3,
            NCloudSchema,
            data_key='simpleSong'
        )

    def current_user_list_playlists(self):
        playlists, _ = self.current_user_playlists()
        return playlists

    def current_user_playlists(self):
        user_id = str(self._user.identifier)
        data_playlists = self.api.user_playlists(user_id)
        playlists = []
        fav_playlists = []
        for pl in data_playlists:
            if str(pl['userId']) == user_id:
                playlists.append(pl)
            else:
                fav_playlists.append(pl)
        return [_deserialize(e, V2PlaylistSchema) for e in playlists], \
            [_deserialize(e, V2PlaylistSchema) for e in fav_playlists]

    def current_user_list_radio_songs(self, count: int = 3) -> List[SongModel]:
        if not self.has_current_user():
            raise NoUserLoggedIn

        if count <= 0:
            return []

        songs_data = self.api.get_radio_music()
        if not songs_data:
            raise NeteaseIOError('failed to fetch radio songs')

        songs = [_deserialize(song_data, V2SongSchema)
                 for song_data in songs_data]
        return songs[:count]

    def rec_list_daily_playlists(self):
        # If no user login, the API return code=301.
        if not self.has_current_user():
            return []

        playlists_data = self.api.get_recommend_playlists()
        rec_playlists = []
        for playlist_data in playlists_data:
            # FIXME: GUI模式下无法显示歌单描述
            playlist_data['coverImgUrl'] = playlist_data['picUrl']
            playlist_data['description'] = None
            playlist = _deserialize(playlist_data, V2PlaylistSchema)
            rec_playlists.append(playlist)
        return rec_playlists

    def rec_list_daily_songs(self):
        songs_data = self.api.get_recommend_songs()
        return [_deserialize(song_data, V2SongSchemaForV3)
                for song_data in songs_data]

    def toplist_list(self):
        return [_deserialize(each, V2PlaylistSchema) for each in self.api.list_toplist()]

    def toplist_get(self, identifier):
        return self.playlist_get(identifier)

    def song_get(self, identifier):
        data = self.api.song_detail(int(identifier))
        return _deserialize(data, V2SongSchema)

    def song_list_similar(self, song):
        songs = self.api.get_similar_song(song.identifier)
        return [_deserialize(song, V2SongSchema) for song in songs]

    def song_get_lyric(self, song):
        data = self.api.get_lyric_by_songid(song.identifier)
        return LyricModel(
            source=SOURCE,
            identifier=self.identifier,
            content=data.get('lrc', {}).get('lyric', ''),
            trans_content=data.get('tlyric', {}).get('lyric', ''),
        )

    def song_get_mv(self, song):
        cache_key = 'mv_id'
        mvid, exists = song.cache_get(cache_key)
        if exists is not True:
            # FIXME: the following implicitly get mv_id attribute
            upgraded_song = self.song_get(song.identifier)
            mvid, exists = upgraded_song.cache_get(cache_key)
            assert exists is True
            song.cache_set(cache_key, mvid)

        if mvid:  # if mvid is valid
            data = self.api.get_mv_detail(mvid)
            mv = _deserialize(data, V2MvSchema)
            return mv
        return None

    def upload_song(self, path: str) -> bool:
        return self.api.cloud_song_upload(path) == 'STATUS_SUCCEEDED'

    def song_list_quality(self, song):
        return list(self._song_get_q_media_mapping(song))

    def song_list_hot_comments(self, song):
        comment_thread_id = self._model_cache_get_or_fetch(song, 'comment_thread_id')
        data = self.api.get_comment(comment_thread_id)
        hot_comments_data = data['hotComments']
        hot_comments = []
        for comment_data in hot_comments_data:
            user_data = comment_data['user']
            user = BriefUserModel(identifier=str(user_data['userId']),
                                  source=SOURCE,
                                  name=user_data['nickname'])
            be_replied = comment_data['beReplied']
            if be_replied:
                replied_comment_data = be_replied[0]
                parent = BriefCommentModel(
                    identifier=replied_comment_data['beRepliedCommentId'],
                    user_name=replied_comment_data['user']['nickname'],
                    content=replied_comment_data['content']
                )
            else:
                parent = None
            comment = CommentModel(identifier=str(comment_data['commentId']),
                                   source=SOURCE,
                                   user=user,
                                   content=comment_data['content'],
                                   liked_count=comment_data['likedCount'],
                                   time=comment_data['time'] // 1000,
                                   parent=parent,
                                   root_comment_id=str(comment_data['parentCommentId']))
            hot_comments.append(comment)
        return hot_comments

    def song_get_media(self, song, quality):
        q_media_mapping = self._song_get_q_media_mapping(song)
        if quality not in q_media_mapping:
            return None
        song_id = int(song.identifier)
        bitrate, url, format = q_media_mapping.get(quality)
        # None means the url is not fetched, so try to fetch it.
        if url is None:
            songs_data = self.api.weapi_songs_url([song_id], bitrate)
            if songs_data:
                song_data = songs_data[0]
                url = song_data['url']
                actual_bitrate = song_data['br']
                format = song_data['type']
                # Check the url bitrate while it is not empty. Api
                # may return a fallback bitrate when the expected bitrate
                # resource is not valid.
                if url and abs(actual_bitrate - bitrate) >= 10000:
                    logger.warning(
                        f'The actual bitrate is {actual_bitrate} '
                        f'while we want {bitrate}. '
                        f'[song:{song_id}].'
                    )
        if url:
            media = Media(url, bitrate=bitrate//1000, format=format)
            # update value in cache
            q_media_mapping[quality] = (bitrate, url, format)
            return media
        logger.error('This should not happend')
        return None

    def _song_get_q_media_mapping(self, song):
        q_media_mapping, exists = song.cache_get('q_media_mapping')
        if exists is True:
            return q_media_mapping

        song_id = int(song.identifier)
        songs_data = self.api.songs_detail_v3([song_id])
        if songs_data:
            q_media_mapping = {}  # {Quality.Audio: (bitrate, url, format)}
            song_data = songs_data[0]
            key_quality_mapping = {
                'h': Quality.Audio.hq,
                'm': Quality.Audio.sq,
                'l': Quality.Audio.lq,
            }

            # Trick: try to find the highest quality url
            # When the song is only for vip/paid user and current user is non-vip,
            # the highest bitrate is 0, which means this song is unavailable
            # for current user.
            songs_url_data = self.api.weapi_songs_url([song_id], 999000)
            assert songs_url_data, 'length should not be 0'
            song_url_data = songs_url_data[0]
            if song_url_data['freeTrialInfo'] is not None:
                logger.debug(f'song:{song_id} only has media segment for trial')
            else:
                highest_bitrate = song_url_data['br']
                # When the bitrate is large than 320000, the quality is treated as
                # lossless. We set the threshold to 400000 here.
                # Note(cosven): From manual testing, the bitrate of lossless media
                # can be 740kbps, 883kbps, 1411kbps, 1777kbps.
                if song_url_data['url'] and 'privatecloud' in song_url_data['url']:
                    # 对于云盘歌曲, netease会抛弃官方音乐地址, 只会返回自己上传的音乐链接
                    # bitrate不由用户提供 由官方估算， 且不再是标准的320, 192, 128
                    q_media_mapping[Quality.Audio.shq] = (highest_bitrate,
                                                          song_url_data['url'],
                                                          song_url_data['type'])
                else:
                    if highest_bitrate > 400000:
                        q_media_mapping[Quality.Audio.shq] = (highest_bitrate,
                                                              song_url_data['url'],
                                                              song_url_data['type'])

                    for key, quality in key_quality_mapping.items():
                        # Ensure the quality info exists.
                        if key in song_data and song_data[key] is not None:
                            # This resource is invalid for current user since the
                            # expected bitrate is large than the highest_bitrate
                            if (song_data[key]['br'] - highest_bitrate) > 10000:
                                continue
                            q_media_mapping[quality] = (song_data[key]['br'], None, None)
        ttl = 60 * 20
        song.cache_set('q_media_mapping', q_media_mapping, ttl)
        return q_media_mapping

    def song_get_web_url(self, song):
        return f'https://music.163.com/#/song?id={song.identifier}'

    def video_get(self, identifier):
        prefix, real_id = identifier.split('_')
        assert prefix == 'mv'
        data = self.api.get_mv_detail(real_id)
        mv = _deserialize(data, V2MvSchema)
        return mv

    def video_get_media(self, video, quality):
        q_media_mapping = self._model_cache_get_or_fetch(video, 'q_media_mapping')
        return q_media_mapping.get(quality)

    def video_list_quality(self, video):
        q_media_mapping = self._model_cache_get_or_fetch(video, 'q_media_mapping')
        return list(q_media_mapping.keys())

    def album_get(self, identifier):
        album_data = self.api.album_infos(identifier)
        if album_data is None:
            raise ModelNotFound
        description = self.api.album_desc(identifier)
        album_data['description'] = description
        album = _deserialize(album_data, V2AlbumSchema)
        return album

    def album_create_songs_rd(self, album):
        album_with_songs = self.album_get(album.identifier)
        return create_reader(album_with_songs.songs)

    def artist_get(self, identifier):
        artist_data = self.api.artist_infos(identifier)
        artist = artist_data['artist']
        artist['songs'] = artist_data['hotSongs'] or []
        description = self.api.artist_desc(identifier)
        artist['description'] = description
        artist['aliases'] = []
        model = _deserialize(artist, V2ArtistSchema)
        return model

    # TODO: artist create albums g
    # TODO: artist create songs g
    def artist_create_songs_rd(self, artist):
        data = self.api.artist_songs(artist.identifier, limit=0)
        count = int(data['total'])

        def g():
            offset = 0
            per = 50
            while offset < count:
                data = self.api.artist_songs(artist.identifier, offset, per)
                for song_data in data['songs']:
                    yield _deserialize(song_data, V2SongSchemaForV3)
                    # In reality, len(data['songs']) may smaller than per,
                    # which is a bug of netease server side, so we set
                    # offset to `offset + per` here.
                offset += per
                per = 100

        return SequentialReader(g(), count)

    def artist_create_albums_rd(self, artist):

        def g():
            data = self.api.artist_albums(artist.identifier)
            if data['code'] != 200:
                yield from ()
            else:
                cur = 1
                while True:
                    for album in data['hotAlbums']:
                        # the songs field will always be an empty list,
                        # we set it to None
                        album['songs'] = None
                        yield _deserialize(album, V2AlbumSchema)
                        cur += 1
                    if data['more']:
                        data = self.api.artist_albums(artist.identifier, offset=cur)
                    else:
                        break

        return create_reader(g())

    def playlist_get(self, identifier):
        data = self.api.playlist_detail_v3(identifier, limit=0)
        playlist = _deserialize(data, V2PlaylistSchema)
        return playlist

    def playlist_delete(self, identifier):
        try:
            self.api.delete_playlist(identifier)
        except CodeShouldBe200 as e:
            logger.warning(f'delete playlist failed, {e}')
            return False
        return True

    def playlist_create_by_name(self, name):
        """
        :raises NoUserLoggedIn:
        :raises CodeShouldBe200: create playlist failed
        """
        user = self.get_current_user()
        uid = user.identifier
        data = self.api.new_playlist(uid, name)
        return _deserialize(data, V2PlaylistSchema)

    def djradio_create_songs_rd(self, djradio_id):
        data = self.api.djradio_list(djradio_id, limit=1, offset=0)
        count = data.get('count', 0)

        def g():
            offset = 0
            per = 50  # speed up first request
            while offset < count:
                tracks_data = self.api.djradio_list(
                    djradio_id, limit=per, offset=offset)
                for track_data in tracks_data.get('programs', []):
                    yield _deserialize(track_data, NDjradioSchema)
                offset += per
        return create_reader(g())

    def playlist_create_songs_rd(self, playlist):
        if playlist.identifier.startswith(DjradioPrefix):
            return self.djradio_create_songs_rd(playlist.identifier[len(DjradioPrefix):])

        data = self.api.playlist_detail_v3(playlist.identifier, limit=0)
        track_ids = data['trackIds']  # [{'id': 1, 'v': 1}, ...]
        count = len(track_ids)

        def g():
            offset = 0
            # 第一次请求应该尽可能快一点，所以这里只获取少量的歌曲。
            # 综合页面展示以及请求速度，拍脑袋将值设置为 50。
            per = 50
            while offset < count:
                end = min(offset + per, count)
                if end <= offset:
                    break
                ids = [track_id['id'] for track_id in track_ids[offset: end]]
                # NOTE(cosven): 记忆中这里有个坑，传入的 id 个数和返回的歌曲个数
                # 不一定相等。比如在一个叫做“万首歌单”的歌单里面，有的歌曲是
                # 获取不到信息的。这也是这里为什么用 SequentialReader 而不是
                # RandomSequentialReader 的原因。
                tracks_data = self.api.songs_detail_v3(ids)
                for track_data in tracks_data:
                    yield _deserialize(track_data, V2SongSchemaForV3)
                offset += per
                # 这里设置为 800 主要是为 readall 的场景考虑的。假设这个值设置很小，
                # 那当这个歌单歌曲数量比较多的时候（比如一万首），需要很多个请求
                # 才能获取到全部的歌曲。
                per = 800

        return SequentialReader(g(), count)

    def playlist_remove_song(self, playlist, song):
        song_id = song.identifier
        rv = self.api.op_music_to_playlist(song_id, playlist.identifier, 'del')
        if rv != 1:
            return False
        return True

    def playlist_add_song(self, playlist, song):
        song_id = song.identifier
        rv = self.api.op_music_to_playlist(song_id, playlist.identifier, 'add')
        if rv == 1:
            song = provider.song_get(song_id)
            return True
        elif rv == -1:
            return True
        return False

    def search(self, keyword, type_, **kwargs):
        type_ = SearchType.parse(type_)
        type_type_map = {
            SearchType.so: 1,
            SearchType.al: 10,
            SearchType.ar: 100,
            SearchType.pl: 1000,
        }
        data = provider.api.search(keyword, stype=type_type_map[type_])
        data['q'] = keyword
        result = _deserialize(data, NeteaseSearchSchema)
        return result

    def get_user_from_cookies(self, cookies):
        """Get a user model with cookies
        """
        if not cookies:  # is None or empty
            raise ValueError('empty cookies')

        provider.api.load_cookies(cookies)
        try:
            data = provider.api.user_level()
        except CodeShouldBe200 as e:
            raise ValueError(f'get user info with cookies failed: {e}')
        else:
            user_id = data['userId']
            user = self.user_get(user_id)
            user.cache_set('cookies', cookies)
            return user


provider = NeteaseProvider()


from .models import _deserialize, create_g, create_cloud_songs_g  # noqa
from .schemas import (  # noqa
    V2SongSchema,
    V2SongSchemaForV3,
    V2MvSchema,
    V2AlbumSchema,
    V2ArtistSchema,
    V2BriefArtistSchema,
    V2BriefAlbumSchema,
    V2PlaylistSchema,
    NeteaseDjradioSchema,
    NeteaseSearchSchema,
    NDjradioSchema,
    NCloudSchema,
    DjradioPrefix,
)



import logging

from feeluown.utils.reader import RandomSequentialReader

from .excs import NeteaseIOError

logger = logging.getLogger(__name__)


def _deserialize(data, schema_cls):
    schema = schema_cls()
    obj = schema.load(data)
    return obj


def create_g(func, schema=None, data_field=None):
    data = func(limit=0)
    if data is None:
        raise NeteaseIOError('server responses with error status code')
    if data_field is None:
        data_field = 'data'

    count = int(data['count'])

    def read_func(start, end):
        data = func(start, end - start)
        return [_deserialize(data, schema)
                for data in data[data_field]]

    reader = RandomSequentialReader(count,
                                    read_func=read_func,
                                    max_per_read=200)
    return reader


def create_cloud_songs_g(func, func_private, schema=None, schema_private=None,
                         data_field=None, data_key=None):
    data = func(limit=0)
    if data is None:
        raise NeteaseIOError('server responses with error status code')
    if data_field is None:
        data_field = 'data'

    count = int(data['count'])

    def read_func(start, end):
        songs_data = func(start, end - start)

        songs = []
        private_songs_info = []
        for idx, song_data in enumerate(songs_data[data_field]):
            if data_key:
                song_data = song_data[data_key]
            if song_data['id'] == song_data['s_id']:
                # FIXME: 有些云盘歌曲在 netease 上不存在，这时不能把它们转换成
                # SongModel。因为 SongModel 的逻辑没有考虑 song 不存在的情况。
                name = song_data['name']
                logger.warning(f'cloud song:{name} may not exist on netease.')
                private_songs_info.append((idx, song_data['id']))
            else:
                song = _deserialize(song_data, schema)
                songs.append(song)

        if private_songs_info:
            private_song_ids = ','.join([str(id) for (_, id) in private_songs_info])
            private_songs_data = func_private(private_song_ids)
            for idx, song_data in enumerate(private_songs_data[data_field]):
                song = _deserialize(song_data, schema_private)
                songs.insert(private_songs_info[idx][0], song)

        return songs

    reader = RandomSequentialReader(count,
                                    read_func=read_func,
                                    max_per_read=200)
    return reader



import logging
import os
from typing import TYPE_CHECKING, Protocol

from feeluown.utils.dispatch import Signal
from feeluown.gui.provider_ui import (
    AbstractProviderUi,
    UISupportsLoginOrGoHome,
    NavBtn,
)
from feeluown.gui.widgets.login import (
    CookiesLoginDialog as _CookiesLoginDialog, InvalidCookies,
)

from .provider import provider
from .login_controller import LoginController

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


class UISupports(UISupportsLoginOrGoHome, Protocol):
    ...


class NeteaseProviderUI(AbstractProviderUi):
    """
    FIXME: 简化 login_as 和 ready_to_login 两个方法的实现逻辑
    """

    def __init__(self, app: 'GuiApp'):
        self._app = app
        self._user = None
        self._login_event = Signal()

    def _(self) -> UISupports:
        return self

    @property
    def login_event(self):
        return self._login_event

    @property
    def provider(self):
        return provider

    def get_colorful_svg(self) -> str:
        return os.path.join(os.path.dirname(__file__), 'assets', 'icon.svg')

    def register_pages(self, route):
        from .page_fav import render as fav_render  # noqa

        route('/providers/netease/fav')(fav_render)

    def login_or_go_home(self):
        if self._user is not None:
            logger.debug('You have already logined in.')
            self.login_event.emit(self, 2)
            return

        logger.debug('Trying to load last login user...')
        user = LoginController.load()
        if user is not None:
            cookies, exists = user.cache_get('cookies')
            assert exists
            if 'MUSIC_U' in cookies:
                logger.debug('Trying to load last login user...done')
                self.provider.auth(user)
                self.login_event.emit(self, 1)
                return

        logger.debug('Trying to load last login user...failed')
        self._dialog = CookiesLoginDialog('https://music.163.com', ['MUSIC_U'])
        self._dialog.login_succeed.connect(self.on_login_succeed)
        self._dialog.show()
        self._dialog.autologin()

    def on_login_succeed(self):
        del self._dialog
        self.login_event.emit(self, 1)

    def list_nav_btns(self):
        return [
            NavBtn(
                icon='☁️',
                text='云盘歌曲',
                cb=lambda: self._app.browser.goto(page='/providers/netease/fav')
            ),
        ]


class CookiesLoginDialog(_CookiesLoginDialog):

    def setup_user(self, user):
        provider.auth(user)

    async def user_from_cookies(self, cookies):
        try:
            user = provider.get_user_from_cookies(cookies)
        except ValueError as e:
            raise InvalidCookies(str(e))
        return user

    def load_user_cookies(self):
        user = LoginController.load()
        if user is not None:
            cookies, exists = user.cache_get('cookies')
            assert exists
            return cookies
        return None

    def dump_user_cookies(self, user, cookies):
        LoginController.save(user)
